/*================================================================================================*
 *      Flash  
 *================================================================================================*/

#include "flash.h"
#include "fastcrc.h"
#include "wdt.h"
#include <intrinsics.h>
#include <string.h>

/*================================================================================================*/
#define FLASH_FWPB_SIGNATURE    0x42505746  /*   , "FWPB" */
/*================================================================================================*/

/*================================================================================================*/
/*  ,    4  */
#pragma pack(4)
typedef struct
    {
    unsigned long Signature;
    t_FLASH_Metadata Metadata;
    unsigned long CRC32;
    }
    t_FLASH_FWPB;
#pragma pack()
/*================================================================================================*/

/*================================================================================================*/
#pragma diag_suppress=Pe237 /*  warning "selector expression is constant" */
#define ASSERT(expr)    switch (0) { case 0: case (expr): ; }
/*================================================================================================*/

/*================================================================================================*/
#pragma location = "FWPARAMS_SECTION"
static __no_init __root const t_FLASH_FWPB f_FWPB;
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
#pragma diag_suppress=Ta021,Ta023
static __ramfunc void f_ProgramFlash
    (unsigned long LinearAddress)
    {
    __disable_interrupt();
    WDT_Restart();
    while (!(AT91C_BASE_MC->MC_FSR & AT91C_MC_FRDY));
    AT91C_BASE_MC->MC_FCR = AT91C_MC_CORRECT_KEY | AT91C_MC_FCMD_START_PROG |
        (AT91C_MC_PAGEN & LinearAddress);
    while (!(AT91C_BASE_MC->MC_FSR & AT91C_MC_FRDY));
    __enable_interrupt();
    }
#pragma diag_default=Ta021,Ta023
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int FLASH_CheckFirmwareState        /*       */
    (
    const t_FLASH_Metadata **pMeta  /* () -     */
    )                               /*  1,   ,  0 */
    {
    unsigned long crc;

    /*     */
    ASSERT(sizeof(unsigned long) == 4)
    ASSERT(sizeof(t_FLASH_FWPB) <= FLASH_PAGE_SIZE)

    if (pMeta)
        *pMeta = &f_FWPB.Metadata;
    /*       */
    if (f_FWPB.Signature != FLASH_FWPB_SIGNATURE)
        return 0;
    crc = CRC32_Calc(0, (unsigned char*)&f_FWPB.Metadata, sizeof(t_FLASH_Metadata));
    if (f_FWPB.CRC32 != crc)
        return 0;
    /*    */
    if (f_FWPB.Metadata.FirmwareSize > FLASH_MAX_FW_SIZE) /*     */
        return 0;
    WDT_Restart();
    /*    CRC32    (240 )   75  */
    crc = CRC32_Calc(0, (unsigned char*)FLASH_FIRMWARE_START, f_FWPB.Metadata.FirmwareSize);
    return (f_FWPB.Metadata.FirmwareCRC32 == crc);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int FLASH_WriteFirmwareBlock        /*     */
    (
    unsigned long FirmwareOffset,   /* ()  -   ,  FLASH_PAGE_SIZE */
    const unsigned long *pData,     /* ()  -    ( 4!) */
    unsigned int BlockSize          /* ()  -    ,  FLASH_PAGE_SIZE */
    )                               /*  1  , 0 =   */
    {
    unsigned long *FlashPtr;

    if ((FirmwareOffset >= FLASH_MAX_FW_SIZE) || (FirmwareOffset % FLASH_PAGE_SIZE))
        return 0;
    if ((!BlockSize) || (BlockSize % FLASH_PAGE_SIZE))
        return 0;
    if (FirmwareOffset + BlockSize > FLASH_MAX_FW_SIZE)
        return 0;
    if ((unsigned long)pData & 3) /*   */
        return 0;
    /*  flash */
    FlashPtr = (void*)(FLASH_FIRMWARE_START + FirmwareOffset);
    while (BlockSize > 0)
        {
        unsigned int k;
        for (k = 0; k < FLASH_PAGE_SIZE / 4; k++)
            {
            *FlashPtr++ = *pData++;
            }
        f_ProgramFlash((unsigned long)FlashPtr - FLASH_PAGE_SIZE);
        BlockSize -= FLASH_PAGE_SIZE;
        }
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int FLASH_WriteMetadata             /*    */
    (
    const t_FLASH_Metadata *pMeta   /* ()  -   */
    )                               /*  1  , 0 =   */
    {
    int k;
    const unsigned long *psrc;
    unsigned long *pdest;

    if ((unsigned int)&f_FWPB % FLASH_PAGE_SIZE)
        return 0;
    if ((!pMeta) || ((unsigned long)pMeta & 3))
        return 0;

    /*    flash *32- * */
    psrc = (const unsigned long*)pMeta;
    pdest = (unsigned long*)&f_FWPB;

    /*   */
    *pdest++ = FLASH_FWPB_SIGNATURE;
    /*   */
    for (k = 0; k < sizeof(t_FLASH_Metadata); k += 4)
        {
        *pdest++ = *psrc++;
        }
    /*    */
    *pdest++ = CRC32_Calc(0, (unsigned char*)pMeta, sizeof(t_FLASH_Metadata));

    /*    FF' */
    while ((unsigned long)pdest % FLASH_PAGE_SIZE)
        {
        *pdest++ = 0xFFFFFFFFUL;
        }
    /*  flash (    RAM) */
    f_ProgramFlash((unsigned long)&f_FWPB);
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
