#include "USBHardware.h"
#include "flash.h"
#include "main.h"
#include "timers.h"

/*================================================================================================*/
#define INIT_SIGNATURE      0x54494E49UL  /*    INIT_SESSION ("INIT") */

/*       (  VDRQ_...) */
#define DEVICE_TO_HOST()    \
    ((CurrentUSBRequest.bmRequestType & REQUEST_ANSWER_DIRECTION) == REQUEST_ANSWER_DIRECTION_DEVICE_TO_HOST)
#define HOST_TO_DEVICE()    \
    ((CurrentUSBRequest.bmRequestType & REQUEST_ANSWER_DIRECTION) != REQUEST_ANSWER_DIRECTION_DEVICE_TO_HOST)

/*   ,   USB. Bit 8 = direction, bits 0..7 = command byte */
#define CMD_INIT_SESSION    0x000   /* (R)       */
#define CMD_GET_INFO        0x001   /* (R)     */
#define CMD_CHECK_FIRMWARE  0x002   /* (R)    */
#define CMD_READ_FIRMWARE   0x003   /* (R)   flash- */
#define CMD_OPEN_SESSION    0x100   /* (W)   ( ) */
#define CMD_WRITE_FIRMWARE  0x101   /* (W)   flash- */
#define CMD_WRITE_METADATA  0x102   /* (W)      flash */
#define CMD_CLOSE_SESSION   0x103   /* (W)   */

/*   */
#define SST_CLOSED          0       /*   ,    INIT_SESSION */
#define SST_INIT            1       /*  ;   SESSION_CONTROL */
#define SST_OPEN            2       /*  ,    */
/*================================================================================================*/

/*================================================================================================*/
/*      USB control channel */
void VDRQ_application(void);
void VDRQ_bootloader_request(void);

/* ,      Device Request USB */
USB_request_Handler VendorDeviceRequest[VDRQ_COUNT] =
    {
    /* 00 */ VDRQ_application,
    /* 01 */ VDRQ_application,
    /* 02 */ VDRQ_application,
    /* 03 */ VDRQ_application,
    /* 04 */ VDRQ_application,
    /* 05 */ VDRQ_application,
    /* 06 */ VDRQ_application,
    /* 07 */ VDRQ_application,
    /* 08 */ VDRQ_application,
    /* 09 */ VDRQ_application,
    /* 0A */ VDRQ_application,
    /* 0B */ VDRQ_application,
    /* 0C */ VDRQ_application,
    /* 0D */ VDRQ_application,
    /* 0E */ VDRQ_application,
    /* 0F */ VDRQ_bootloader_request
    };

#pragma pack(1)
/*    ,    GET_INFO */
static const struct
    {
    unsigned short BootromVersion;
    unsigned short FlashPageSize;
    unsigned long MaxFirmwareSize;
    char ModuleName[8];
    }
    f_BootromInfo = { BOOTROM_VERSION, FLASH_PAGE_SIZE, FLASH_MAX_FW_SIZE, "E140" };
/*    INIT_SESSION (   ) */
static struct
    {
    unsigned long Signature;
    unsigned long ControlKey;
    }
    f_SessionInit;
#pragma pack()
static unsigned char f_SessionState = SST_CLOSED;
/*================================================================================================*/

/*================================================================================================*/

/**************************************************************************************************
 ** CALLBACK'     USB control pipe                                    **
 **************************************************************************************************/

/*------------------------------------------------------------------------------------------------*/
int cb_USB_CheckFW
    (unsigned int UserParam, unsigned char *pBuffer, unsigned int RequestLen)
    {
    /*   ,       */
    if (f_SessionState != SST_OPEN)
        return -1;
    if (RequestLen < sizeof(t_FLASH_Metadata))
        return -1;
    g_FW_Valid = FLASH_CheckFirmwareState(&g_FW_pMeta);
    if (g_FW_Valid)
        {
        memcpy(pBuffer, g_FW_pMeta, sizeof(t_FLASH_Metadata));
        }
    else
        {
        memset(pBuffer, 0, sizeof(t_FLASH_Metadata));
        }
    return sizeof(t_FLASH_Metadata);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int cb_USB_SessionClose
    (unsigned int UserParam, unsigned char *pBuffer, unsigned int RequestLen)
    {
    f_SessionState = SST_CLOSED;
    return RequestLen;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int cb_USB_SessionOpen
    (unsigned int UserParam, unsigned char *pBuffer, unsigned int RequestLen)
    {
    unsigned long key;

    if (f_SessionState == SST_CLOSED)
        return -1;
    if (RequestLen != 4)
        return -1;

    key = (unsigned long)pBuffer[0] | ((unsigned long)pBuffer[1] << 8) |
          ((unsigned long)pBuffer[2] << 16) | ((unsigned long)pBuffer[3] << 24);
    if (key != f_SessionInit.ControlKey)
        return -1;
    f_SessionState = SST_OPEN;
    return RequestLen;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int cb_USB_WriteFW
    (unsigned int UserParam, unsigned char *pBuffer, unsigned int RequestLen)
    {
    /* OpCode =    */
    if (f_SessionState != SST_OPEN)
        return -1;
    if (!FLASH_WriteFirmwareBlock(UserParam, (void*)pBuffer, RequestLen))
        return -1;
    return RequestLen;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int cb_USB_WriteMeta
    (unsigned int UserParam, unsigned char *pBuffer, unsigned int RequestLen)
    {
    if (f_SessionState != SST_OPEN)
        return -1;
    if (RequestLen < sizeof(t_FLASH_Metadata))
        return -1;
    if (!FLASH_WriteMetadata((void*)pBuffer))
        return -1;
    return sizeof(t_FLASH_Metadata);
    }
/*------------------------------------------------------------------------------------------------*/

/**************************************************************************************************
 **  Vendor Device Request'                                                         **
 **************************************************************************************************/

/*------------------------------------------------------------------------------------------------*/
void VDRQ_application(void)
    {
    g_MainFlag |= MAIN_RUN_USER_FIRMWARE;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
/*     bootload,        bootloader */
void VDRQ_bootloader_request(void)
    {
    unsigned int cmd, FWOffset;

    cmd = (CurrentUSBRequest.wIndex >> 8) & 0xFF;
    FWOffset = ((CurrentUSBRequest.wIndex & 0xFF) << 16) | CurrentUSBRequest.wValue;
    if (HOST_TO_DEVICE())
        cmd |= 0x100; /*   */
    switch (cmd)
        {
        case CMD_INIT_SESSION:
            /* (R)        */
            f_SessionState = SST_INIT;
            f_SessionInit.Signature = INIT_SIGNATURE;
            f_SessionInit.ControlKey = g_TIMER_PseudoRandom;
            if (!USBBG_BeginWithBuffer(USBBG_READ, 0, CurrentUSBRequest.wLength, NULL,
                                       (void*)&f_SessionInit, sizeof(f_SessionInit)))
                USBBG_RequestError();
            break;
        case CMD_GET_INFO:
            /* (R)     */
            if (f_SessionState != SST_OPEN)
                USBBG_RequestError();
            if (!USBBG_BeginWithBuffer(USBBG_READ, 0, CurrentUSBRequest.wLength, NULL,
                                       (void*)&f_BootromInfo, sizeof(f_BootromInfo)))
                USBBG_RequestError();
            break;
        case CMD_READ_FIRMWARE:
            /* (R)   */
            if (f_SessionState != SST_OPEN)
                USBBG_RequestError();
            if (!USBBG_BeginWithBuffer(USBBG_READ, 0, CurrentUSBRequest.wLength, NULL,
                                       (void*)(FLASH_FIRMWARE_START + FWOffset),
                                       CurrentUSBRequest.wLength))
                USBBG_RequestError();
            break;
        case CMD_CHECK_FIRMWARE:
            /* (R)    */
            if (!USBBG_Begin(USBBG_READ, 0, CurrentUSBRequest.wLength, cb_USB_CheckFW))
                USBBG_RequestError();
            break;
        case CMD_OPEN_SESSION:
            /* (W)   ( ) */
            if (!USBBG_Begin(USBBG_WRITE, 0, CurrentUSBRequest.wLength, cb_USB_SessionOpen))
                USBBG_RequestError();
            break;
        case CMD_CLOSE_SESSION:
            /* (W)   */
            if (!USBBG_Begin(USBBG_WRITE, 0, CurrentUSBRequest.wLength, cb_USB_SessionClose))
                USBBG_RequestError();
            break;
        case CMD_WRITE_FIRMWARE:
            /* (W)   flash- */
            if (!USBBG_Begin(USBBG_WRITE, FWOffset, CurrentUSBRequest.wLength, cb_USB_WriteFW))
                USBBG_RequestError();
            break;
        case CMD_WRITE_METADATA:
            /* (W)      flash */
            if (!USBBG_Begin(USBBG_WRITE, 0, CurrentUSBRequest.wLength, cb_USB_WriteMeta))
                USBBG_RequestError();
            break;
        default:
            USBBG_RequestError();
            break;
        }
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
