/*================================================================================================*
 *    E-140
 *================================================================================================*/

#include "dac.h"
#include "cpld.h"
#include "timers.h"
#include "chipio.h"
#include "spi.h"
#include "USBHardware.h"
#include "debug.h"
#include "pio.h"

/*================================================================================================*/
/*   */
#define DAC_ST_STOPPED      0       /*   */
#define DAC_ST_STREAMING    1       /*     ( DMA) */
#define DAC_ST_START_REQ    2       /*        */

/*      N */
#define RING_NEXT(idx, N)  (((idx) < ((N) - 1)) ? ((idx) + 1) : 0)

/*   ()   */
#define BS_EMPTY            0       /*  */
#define BS_FULL             1       /*   */

/*     */
#define DAC_UNDERRUN_ZERO   1       /* 1 =    ; 0 =   */

#define DAC_DMA_SIZE        (DAC_BLK_SIZE / 2) /*   16-  */
/*================================================================================================*/

/*================================================================================================*/
/*       */
static unsigned int f_SendIdx = 0;  /*    */
static unsigned int f_LoadIdx = 0;  /*      */
static unsigned int f_BufState[DAC_NBLOCKS];
/*   = 32 ,   16-  (    ) */
static volatile unsigned int f_Buffer[DAC_NBLOCKS][DAC_BLK_NSAMPLES];
#if DAC_UNDERRUN_ZERO
static unsigned int f_ZeroBuffer[DAC_BLK_NSAMPLES]; /*    */
#endif

/*        */
static int f_RunState = DAC_ST_STOPPED;

/*   ( ,   ) */
static t_DAC_ParamStruct f_DAC_Params;
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
__arm __ramfunc void SPI_Interrupt_Handler
    (void)
    {
    unsigned int status;

    status = AT91C_BASE_SPI->SPI_SR;
    status &= AT91C_BASE_SPI->SPI_IMR;

    if (f_RunState != DAC_ST_STREAMING)
        {
        AT91C_BASE_SPI->SPI_IDR = 0xFFFFFFFF;
        return;
        }

    if (status & AT91C_SPI_ENDTX)
        {
        /*   ,    */
        /*    ,  DMA    
         *   f_SendIdx       ( 
         *   TNPR/TNCR).
         *
         *   "f_SendIdx + 2" (  ).   , 
         *     DMA,   f_SendIdx  
         *     f_SendIdx  1 .
         *    "+2" ,     .   
         * DAC_UNDERRUN_ZERO     ,    
         * .  f_SendIdx     .
         */
        unsigned int plus1, plus2;

        plus1 = RING_NEXT(f_SendIdx, DAC_NBLOCKS);
        plus2 = RING_NEXT(plus1, DAC_NBLOCKS);

        if (f_BufState[plus2] == BS_FULL)
            {
#if DEBUG_DAC_LED_WHEN_UNDERRUN
            AT91C_BASE_PIOA->PIO_CODR = PIO_LED;
#endif
            f_BufState[f_SendIdx] = BS_EMPTY;
            f_SendIdx = plus1;
            AT91C_BASE_SPI->SPI_TNPR = (unsigned int)f_Buffer[plus2];
            }
        else
            {
#if DEBUG_DAC_LED_WHEN_UNDERRUN
            AT91C_BASE_PIOA->PIO_SODR = PIO_LED;
#endif
#if DAC_UNDERRUN_ZERO
            AT91C_BASE_SPI->SPI_TNPR = (unsigned int)f_ZeroBuffer;
#else
            AT91C_BASE_SPI->SPI_TNPR = (unsigned int)f_Buffer[plus1];
#endif
            }
        AT91C_BASE_SPI->SPI_TNCR = DAC_DMA_SIZE;
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_ClearBuffer
    (void)
    {
    /*       */
    unsigned int k;

    /*    */
    __disable_interrupt();
    f_SendIdx = 0;
    f_LoadIdx = 0;
    for (k = 0; k < DAC_NBLOCKS; k++)
        {
        f_BufState[k] = BS_EMPTY;
        }
    __enable_interrupt();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void DAC_Init                       /*    */
    (void)
    {
#if DAC_UNDERRUN_ZERO
    memset(f_ZeroBuffer, 0, sizeof(f_ZeroBuffer));
#endif
    f_RunState = DAC_ST_STOPPED;
    f_ClearBuffer();
    f_DAC_Params.RateDiv = 0;
    f_DAC_Params.SyncWithADC = 0;
    f_DAC_Params.SetZeroOnStop = 1;
    /*    */
    SetInterrupt(AT91C_ID_SPI, SPI_Interrupt_Handler, PRI_HIGH, INTERNAL_HIGH_LEVEL);
    EnableInterruptsMask(1 << AT91C_ID_SPI);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void DAC_GetParams                  /*    ( ) */
    (
    t_DAC_ParamStruct *pParams
    )
    {
    if (pParams)
        memcpy(pParams, &f_DAC_Params, sizeof(f_DAC_Params));
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void DAC_SetParams                  /*    ( ) */
    (
    const t_DAC_ParamStruct *pParams
    )
    {
    if (pParams)
        memcpy(&f_DAC_Params, pParams, sizeof(f_DAC_Params));
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int DAC_SingleSample                /*     */
    (
    unsigned int DAC_Val            /*  16  =  0,  16  =  1 */
    )                               /*  1  , 0   */
    {
    int ok;
    volatile unsigned int dac_data;

    /*      ,      */
    if (f_RunState != DAC_ST_STOPPED)
        return 0;

    if (!CPLD_PowerGood())
        return 0;

    /*  ,       DMA */
    dac_data = DAC_Val;

    /*   */
    if (!SPI_OpenForDAC())
        return 0;

    /*  DMA     dac_data */
    AT91C_BASE_SPI->SPI_TPR = (unsigned int)&dac_data;
    AT91C_BASE_SPI->SPI_TCR = 2; /*   16-  */
    AT91C_BASE_SPI->SPI_TNPR = (unsigned int)&dac_data;
    AT91C_BASE_SPI->SPI_TNCR = 2; /*   16-  */
    /*    DMA,    */
    SPI_StartDMATransmit(0);

    ok = 0;
    do
        {
        if (!CPLD_CFG_SetConfigBits(
            (CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE),
            (CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE)))
            break;
        /*   -    */
        if (!CPLD_DAC_Async(CPLD_DACCTL_ASYNC_PRELOAD))
            break;
        /*   -         */
        if (!CPLD_DAC_Async(CPLD_DACCTL_ASYNC_OUTPUT))
            break;
        ok = 1;
        }
    while (0);

    /*    SPI    SPI */
    CPLD_CFG_SetConfigBits(CPLD_CONFIG_SPI_MASTER_ENABLE, 0);
    SPI_Close();

    return ok;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int DAC_StartStream                 /*      */
    (void)                          /*  1  , 0   */
    {
    if ((f_RunState != DAC_ST_STOPPED) && (f_RunState != DAC_ST_START_REQ))
        return 0;
    f_RunState = DAC_ST_START_REQ;
    if (!CPLD_PowerGood())
        return 0;
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int DAC_Stop                        /*        */
    (void)                          /*  1  , 0   */
    {
    /*    ,      */
    if (f_RunState != DAC_ST_STREAMING)
        {
        /*    START_REQ,  ( ) */
        f_RunState = DAC_ST_STOPPED;
        /*   */
        f_ClearBuffer();
        /*  endpoint */
        ResetBulkEndpoint(OUT_ENDPOINT);
        return 1;
        }

    /* : f_RunState == DAC_ST_STREAMING */

    /*   */
    if (!CPLD_DAC_Stop())
        return 0;
    f_RunState = DAC_ST_STOPPED;
    SPI_Close();
    if (f_DAC_Params.SetZeroOnStop)
        /*      */
        DAC_SingleSample(0);
    /*    SPI  */
    CPLD_CFG_SetConfigBits(CPLD_CONFIG_SPI_MASTER_ENABLE, 0);
    f_ClearBuffer();
    /*  endpoint */
    ResetBulkEndpoint(OUT_ENDPOINT);

#if DEBUG_DAC_LED_WHEN_RUNNING
    AT91C_BASE_PIOA->PIO_CODR = PIO_LED;
#endif
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void DAC_Worker                     /*    */
    (void)
    {
/*  :
 *  (1)     
 *          USB    .
 *
 *  (2)         (  2 )
 *   SPI      (  slave mode).
 *            .
 *           .
 */
    int ok, k;

    if (f_BufState[f_LoadIdx] == BS_EMPTY)
        {
#if DEBUG_TEST_DAC
        for (k = 0; k < DAC_BLK_NSAMPLES; k++)
            {
            f_Buffer[f_LoadIdx][k] = (k & 1) ? 0x90006FFF : 0x80007FFF;
            }
        f_BufState[f_LoadIdx] = BS_FULL;
        f_LoadIdx = RING_NEXT(f_LoadIdx, DAC_NBLOCKS);
#else
        if (USB_OUT_ArrayCounter >= DAC_BLK_SIZE)
            {
            /*    */
            ReadUSB((void*)f_Buffer[f_LoadIdx], DAC_BLK_SIZE);
            /*     */
            f_BufState[f_LoadIdx] = BS_FULL;
            /*   */
            f_LoadIdx = RING_NEXT(f_LoadIdx, DAC_NBLOCKS);
            }
#endif
        }

    /*     ,      */
    if (f_RunState != DAC_ST_START_REQ)
        return;

    /*   ,   */
    if (!CPLD_PowerGood())
        {
        f_RunState = DAC_ST_STOPPED;
        return;
        }

    /*    DAC_FULL_BLOCKS_BEFORE_START   */
    for (k = 0; k < DAC_FULL_BLOCKS_BEFORE_START; k++)
        {
        if (f_BufState[k] != BS_FULL)
            return;
        }

    if (!SPI_OpenForDAC())
        {
        f_RunState = DAC_ST_STOPPED;
        return;
        }

    f_RunState = DAC_ST_STREAMING;
    f_SendIdx = 0;

    AT91C_BASE_SPI->SPI_TPR = (unsigned int)f_Buffer[0];
    AT91C_BASE_SPI->SPI_TCR = DAC_DMA_SIZE;
    AT91C_BASE_SPI->SPI_TNPR = (unsigned int)f_Buffer[1];
    AT91C_BASE_SPI->SPI_TNCR = DAC_DMA_SIZE;

    SPI_StartDMATransmit(1);

    ok = CPLD_CFG_SetConfigBits(
         CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE,
         CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE);
    /*    */
    if (ok)
        ok = CPLD_DAC_Async(CPLD_DACCTL_ASYNC_PRELOAD);
    if (ok)
        ok = CPLD_DAC_Start(f_DAC_Params.SyncWithADC, f_DAC_Params.RateDiv);
    if (!ok)
        { /*      SPI ,      SPI */
        CPLD_CFG_SetConfigBits((CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE), 0);
        SPI_Close();
        f_RunState = DAC_ST_STOPPED;
        return;
        }
#if DEBUG_DAC_LED_WHEN_RUNNING
    AT91C_BASE_PIOA->PIO_SODR = PIO_LED;
#endif
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
