/*================================================================================================*
 *    E-140
 *================================================================================================*/

#include "dac.h"
#include "buffers.h"
#include "cpld.h"
#include "timers.h"
#include "chipio.h"
#include "spi.h"
#include "main.h"
#include "USBHardware.h"
#include "debug.h"
#include "pio.h"
#include <stddef.h>

/*================================================================================================*/
/*   */
#define DAC_ST_STOPPED      0       /*   */
#define DAC_ST_STREAMING    1       /*     ( DMA) */
#define DAC_ST_START_REQ    2       /*        */

/*      N */
#define RING_NEXT(idx, N)  (((idx) < ((N) - 1)) ? ((idx) + 1) : 0)

/*   ()   */
#define BS_EMPTY            0       /*  */
#define BS_FULL             1       /*   */

/*     */
#define DAC_UNDERRUN_ZERO   1       /* 1 =    ; 0 =   */
/*================================================================================================*/

/*================================================================================================*/
/*       */
static unsigned int f_SendIdx = 0;  /*    */
static unsigned int f_LoadIdx = 0;  /*      */
static volatile unsigned char f_BufState[DAC_NBLOCKS];
#if DAC_UNDERRUN_ZERO
static const unsigned int ZeroBuffer[DAC_BLK_SIZE_DW]; /*    */
#endif

/*        */
static int f_RunState = DAC_ST_STOPPED;

/*   ( ,   ) */
static t_DAC_ParamStruct f_DAC_Params;

/*  ,       */
static unsigned short f_NeedBlocks = 0;

/* --     (  DAC_StartStream()) -- */
static unsigned char f_UseAutoStop = 0;     /*    */
static unsigned char f_UseStopConst = 0;    /*      */
static unsigned int f_StopConst = 0;        /*   (= f_DAC_Params.StopState[]) */
static unsigned char f_AutoStopClrBuf = 0;  /*       */
static unsigned short f_StartOffset = 0;
static unsigned short f_saved_RateDiv;
static unsigned char f_saved_SyncWithADC;

/*         32-  */
/*  0,     */
static unsigned int f_LoopSamples = 0;
/*     */
static unsigned int f_SamplesLeft = 0;
/*    (0 = ) */
static unsigned char f_AutoStopDownCount;

/*    ( f_DAC_Params.StopState[])      */
/*    ,        
          ( ).
        ,      ,
    ,        DMA.
*/
static volatile unsigned int f_StopBuf[DAC_MIN_DMA_SAMPLES * 2 - 1];
/*    (max DAC_MIN_DMA_SAMPLES - 1),     f_StopBuf */
static unsigned int f_StopTailLen = 0;
/*   ,      */
static volatile unsigned int* f_StopRepeatPtr = f_StopBuf;
/*        */
static volatile unsigned int* f_LoopStart = g_Buf.DAC.asFrame;
/*    DMA,      */
static volatile unsigned int* f_LastDMAPtr;

/*     : */
static unsigned int f_LastCountCur;     /*      TCR */
static unsigned int f_LastCountNext;    /*      TNCR */
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
#pragma diag_suppress = Ta022 //  warning possible ROM access from ramfunc
__arm __ramfunc void SPI_Interrupt_Handler
    (void)
    {
    unsigned int status;

    status = AT91C_BASE_SPI->SPI_SR;
    status &= AT91C_BASE_SPI->SPI_IMR;

    if (f_RunState != DAC_ST_STREAMING)
        {
        AT91C_BASE_SPI->SPI_IDR = 0xFFFFFFFF;
        return;
        }

#if DEBUG_LED_DAC_INTERRUPT
    DEBUG_LED_ON();
#endif

    /* :     (   ,  ) */
    if (status & AT91C_SPI_ENDTX)
        {
        volatile unsigned int* next_ptr;
        unsigned int chunk_size = 0;
        unsigned int underrun = 0; /*     (  ) */

        /*    ()     */
        g_Diag.DAC_SampleCount += f_LastCountCur;
        /*       (    DMA  TCR  TNCR) */
        f_LastCountCur = f_LastCountNext;
        f_LastCountNext = 0;

        if (f_UseAutoStop && (f_SamplesLeft == 0))
            { /*   ,    */
            next_ptr = f_StopRepeatPtr;
            chunk_size = DAC_MIN_DMA_SAMPLES;
            if (f_AutoStopDownCount)
                f_AutoStopDownCount--;
            }
        else if (f_LoopSamples)
            { /* ========   ======== */
            f_LastCountNext = chunk_size = f_LoopSamples;
            /*      */
            next_ptr = f_LoopStart;
            /* ...        */
            }
        else
            { /* ========   ======== */

            /*    ,  DMA    
             *   f_SendIdx       ( 
             *   TNPR/TNCR).
             *
             *   "f_SendIdx + 2" (  ).   , 
             *     DMA,   f_SendIdx  
             *     f_SendIdx  1 .
             *    "+2" ,     .   
             * DAC_UNDERRUN_ZERO     ,    
             * .  f_SendIdx     .
             */

            unsigned int plus1, plus2;
            f_LastCountNext = chunk_size = DAC_BLK_SIZE_DW;

            plus1 = RING_NEXT(f_SendIdx, DAC_NBLOCKS);
            plus2 = RING_NEXT(plus1, DAC_NBLOCKS);

            if (f_BufState[plus2] == BS_FULL)
                {
#if DEBUG_LED_DAC_UNDERRUN
                DEBUG_LED_OFF();
#endif
                f_BufState[f_SendIdx] = BS_EMPTY;
                f_SendIdx = plus1;
                next_ptr = g_Buf.DAC.asRing[plus2];
                }
            else
                { /* ,    (    ) */
#if DAC_UNDERRUN_ZERO
                next_ptr = (void*)ZeroBuffer;
#else
                next_ptr = g_Buf.DAC.asRing[plus1];
#endif
                f_LastCountNext = 0; /*       */
                underrun = 1;
                }
            }

        /* ========     ======== */

        if (f_UseAutoStop && (f_SamplesLeft > 0))
            {
            /* f_SamplesLeft    , 
                  (   DMA NEXT  DMA CURRENT) */
            if (f_SamplesLeft <= chunk_size)
                {
                /*    .   -. */

                /*   ,  f_StopBuf   .
                        . */
                if (f_LoopSamples)
                    {
                    next_ptr = f_StopBuf;
                    }
                else
                    {
                    unsigned int stop_sample, k;
                    if (f_StopTailLen)
                        {
                        /*    ,       ,
                                   (next_ptr). */
                        /* :     ,     */
                        if (!underrun)
                            {
                            /*    next_ptr  -. */
                            stop_sample = (f_UseStopConst)
                                ? f_StopConst                      /*   */
                                : next_ptr[f_StopTailLen - 1];     /*     */
                            f_StopRepeatPtr =
                                next_ptr + f_StopTailLen;
                            /* next_ptr    */
                            }
                        }
                    else
                        {
                        /*    ,      
                           (f_SamplesLeft ),      -
                             */
                        underrun = 0; /*   underrun,      */
                        stop_sample = (f_UseStopConst)
                            ? f_StopConst                      /*   */
                            : f_LastDMAPtr[f_SamplesLeft - 1]; /*     */
                        next_ptr = f_StopRepeatPtr = f_StopBuf;
                        }
                    /*  -   */
                    if (!underrun)
                        {
                        for (k = 0; k < DAC_MIN_DMA_SAMPLES; k++)
                            {
                            f_StopRepeatPtr[k] = stop_sample;
                            }
                        }
                    }
                /*      */
                if (!underrun)
                    {
                    f_SamplesLeft = 0;
                    chunk_size = DAC_MIN_DMA_SAMPLES;
                    f_LastCountNext = f_StopTailLen;
                    }
                }
            else
                {
                /*     */
                /*  ,     */
                if (!underrun) /*        */
                    {
                    f_SamplesLeft -= chunk_size; /*  > 0 */
                    if (f_SamplesLeft < chunk_size)
                        f_LastCountNext = chunk_size = f_SamplesLeft;
                    }
                }
            }
        /*     */
        f_LastDMAPtr = next_ptr;
        /*         DMA */
        AT91C_BASE_SPI->SPI_TNPR = (unsigned int)next_ptr;
        AT91C_BASE_SPI->SPI_TNCR = chunk_size * 2;
        /*  ,    */
        if (underrun)
            {
#if DEBUG_LED_DAC_UNDERRUN
            DEBUG_LED_ON();
#endif
            INC_SATURATE(g_Diag.DAC_UnderrunCount);
            }
        }
#if DEBUG_LED_DAC_INTERRUPT
    DEBUG_LED_OFF();
#endif
    }
#pragma diag_default = Ta022
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_ClearBuffer
    (void)
    {
    /*       */
    unsigned int k;

    /*    */
    __disable_interrupt();
    f_SendIdx = 0;
    f_LoadIdx = 0;
    for (k = 0; k < DAC_NBLOCKS; k++)
        {
        f_BufState[k] = BS_EMPTY;
        }
    __enable_interrupt();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void DAC_Init                       /*    */
    (void)
    {
    f_RunState = DAC_ST_STOPPED;
    f_NeedBlocks = 0;
    f_LoopSamples = 0;
    DIAG_CLEAR_FLAG(DIAG_DEVSTATE_DAC_STARTED);
    g_Diag.DAC_UnderrunCount = 0;
    g_Diag.DAC_SampleCount = 0;
    f_ClearBuffer();
    ResetBulkEndpoint(OUT_ENDPOINT); /*  endpoint */
    memset(&f_DAC_Params, 0, sizeof(f_DAC_Params));
    f_DAC_Params.RunMode = DAC_MODE_SET_CONST_ON_STOP;
    f_DAC_Params.PreloadSamples = DAC_DEFAULT_PRELOAD_SAMPLES;
    /*    */
    SetInterrupt(AT91C_ID_SPI, SPI_Interrupt_Handler, PRI_HIGH, INTERNAL_HIGH_LEVEL);
    EnableInterruptsMask(1 << AT91C_ID_SPI);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void DAC_GetParams                  /*    ( ) */
    (
    t_DAC_ParamStruct *pParams
    )
    {
    if (pParams)
        memcpy(pParams, &f_DAC_Params, sizeof(f_DAC_Params));
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int DAC_SetParams                   /*    ( ) */
    (
    const t_DAC_ParamStruct *pParams
    )                               /*  1  , 0   */
    {
    unsigned int preload;

    if (!pParams) return 0;

    /*  PreloadSamples == 0,      */
    preload = pParams->PreloadSamples;
    if (preload == 0)
        preload = DAC_DEFAULT_PRELOAD_SAMPLES;

    /* ====     ==== */

    /*   : */
    /* -  preload      */
    if (preload > DAC_BUFFER_SIZE_DW)
        return 0;

    if (pParams->RunMode & DAC_MODE_LOOP)
        { /*   : */
        /* -    (   ) */
        /* -      */
        if (pParams->StartOffset >= preload)
            return 0;
        /* - total samples   ,  .. 0 */
        }
    else
        { /*   : */
        /* -      - 2  */
        if (preload < (2 * DAC_BLK_SIZE_DW))
            return 0;
        /* -      ,    preload */
        if ((pParams->TotalSamples != 0) && (pParams->TotalSamples < preload))
            return 0;
        }

    /*   */
    memcpy(&f_DAC_Params, pParams, sizeof(f_DAC_Params));
    f_DAC_Params.PreloadSamples = preload;
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int DAC_SingleSample                /*     */
    (
    short int Ch0_Val,              /*   0 */
    short int Ch1_Val               /*   1 */
    )                               /*  1  , 0   */
    {
    int ok;
    volatile short int dac_data[2];

    /*      ,      */
    if (f_RunState != DAC_ST_STOPPED)
        return 0;

    if (!CPLD_PowerGood())
        return 0;

    /*  ,       DMA */
    dac_data[0] = Ch0_Val;
    dac_data[1] = Ch1_Val;

    /*   */
    if (!SPI_OpenForDAC())
        return 0;

    /*  DMA     dac_data */
    AT91C_BASE_SPI->SPI_TPR = (unsigned int)dac_data;
    AT91C_BASE_SPI->SPI_TCR = 2; /*   16-  */
    AT91C_BASE_SPI->SPI_TNPR = (unsigned int)dac_data;
    AT91C_BASE_SPI->SPI_TNCR = 2; /*   16-  */
    /*    DMA,    */
    SPI_StartDMATransmit(0);

    ok = 0;
    do
        {
        if (!CPLD_CFG_SetConfigBits(
            (CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE),
            (CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE)))
            break;
        /*   -    */
        if (!CPLD_DAC_Async(CPLD_DACCTL_ASYNC_PRELOAD))
            break;
        /*   -         */
        if (!CPLD_DAC_Async(CPLD_DACCTL_ASYNC_OUTPUT))
            break;
        ok = 1;
        }
    while (0);

    /*    SPI    SPI */
    CPLD_CFG_SetConfigBits(CPLD_CONFIG_SPI_MASTER_ENABLE, 0);
    SPI_Close();

    return ok;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int DAC_StartStream                 /*      */
    (void)                          /*  1  , 0   */
    {
    if ((f_RunState != DAC_ST_STOPPED) && (f_RunState != DAC_ST_START_REQ))
        return 0;
    if (!CPLD_PowerGood())
        return 0;
    /*   ,     USB */
    f_NeedBlocks = (f_DAC_Params.PreloadSamples + DAC_BLK_SIZE_DW - 1) / DAC_BLK_SIZE_DW;
    /*     .   - 0 */
    f_LoopSamples =
        (f_DAC_Params.RunMode & DAC_MODE_LOOP) ? f_DAC_Params.PreloadSamples : 0;
    /*      (0 = ) */
    f_UseAutoStop = (f_DAC_Params.TotalSamples > 0);
    f_SamplesLeft = f_DAC_Params.TotalSamples;
    /*    (  ) */
    f_StartOffset = f_DAC_Params.StartOffset;
    /*    */
    f_UseStopConst = ((f_DAC_Params.RunMode & DAC_MODE_SET_CONST_ON_STOP) != 0);
    /*     f_DAC_Params    */
    f_StopConst = ((unsigned int)f_DAC_Params.StopState[1] << 16)
                  | (f_DAC_Params.StopState[0] & 0xFFFF);
    f_AutoStopClrBuf =
            ((f_DAC_Params.RunMode & (DAC_MODE_LOOP | DAC_MODE_NO_CLEAR_BUFFER))
                                  != (DAC_MODE_LOOP | DAC_MODE_NO_CLEAR_BUFFER));
    f_saved_SyncWithADC = f_DAC_Params.SyncWithADC;
    f_saved_RateDiv = f_DAC_Params.RateDiv;
    /*       */
    g_Diag.DAC_UnderrunCount = 0;
    g_Diag.DAC_SampleCount = 0;
    /*      DAC_Worker */
    f_RunState = DAC_ST_START_REQ;
    /*  ,    */
    DIAG_SET_FLAG(DIAG_DEVSTATE_DAC_STARTED);
#if DEBUG_LED_DAC_RUNNING
    DEBUG_LED_ON();
#endif
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int DAC_Stop                        /*        */
    (
    unsigned int clear_buf          /* 0 =    */
    )                               /*  1  , 0   */
    {
    switch (f_RunState)
        {
        case DAC_ST_START_REQ:
        case DAC_ST_STOPPED:
            /*    START_REQ,  ( ) */
            f_RunState = DAC_ST_STOPPED;
            DIAG_CLEAR_FLAG(DIAG_DEVSTATE_DAC_STARTED);
            if (clear_buf)
                f_ClearBuffer(); /*   */
#if DEBUG_LED_DAC_RUNNING
            DEBUG_LED_OFF();
#endif
            return 1;
        case DAC_ST_STREAMING:
            /*   */
            if (!CPLD_DAC_Stop())
                return 0;
            f_RunState = DAC_ST_STOPPED;
            DIAG_CLEAR_FLAG(DIAG_DEVSTATE_DAC_STARTED);
            SPI_Close();
            /*    :     DMA */
            {
            int dma_samples_left = (AT91C_BASE_SPI->SPI_TCR / 2) + 1; /* +1 -  */
            __disable_interrupt();
            if (f_LastCountCur > dma_samples_left)
                {
                g_Diag.DAC_SampleCount += (f_LastCountCur - dma_samples_left);
                f_LastCountCur = 0;
                }
            __enable_interrupt();
            }
            /*  ,    */
            if (f_UseStopConst)
                DAC_SingleSample(f_StopConst & 0xFFFF, f_StopConst >> 16);
            /*    SPI  */
            CPLD_CFG_SetConfigBits(CPLD_CONFIG_SPI_MASTER_ENABLE, 0);
            if (clear_buf)
                f_ClearBuffer(); /*   */
#if DEBUG_LED_DAC_RUNNING
            DEBUG_LED_OFF();
#endif
            return 1;
        default:
            return 0;
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void DAC_Worker                     /*    */
    (void)
    {
/*  :
 *  (1)     
 *          USB    .
 *
 *  (2)        
 *   SPI      (  slave mode).
 *            .
 *           .
 *
 *  (3)        (  )
 */
    int ok;

    if (f_BufState[f_LoadIdx] == BS_EMPTY)
        {
        if (USB_OUT_ArrayCounter >= DAC_BLK_SIZE_B)
            {
            /*    */
            ReadUSB((void*)g_Buf.DAC.asRing[f_LoadIdx], DAC_BLK_SIZE_B);
            /*     */
            f_BufState[f_LoadIdx] = BS_FULL;
            /*   */
            f_LoadIdx = RING_NEXT(f_LoadIdx, DAC_NBLOCKS);
            }
        }

    /*     ,   */
    if ((f_RunState == DAC_ST_STREAMING) && (!f_AutoStopDownCount))
        {
        /*   ,       DMA */
        f_UseStopConst = 0;
        /*  ,      DAC_MODE_LOOP  DAC_MODE_NO_CLEAR_BUFFER */
        DAC_Stop(f_AutoStopClrBuf);
        return;
        }

    /*     ,      */
    if (f_RunState != DAC_ST_START_REQ)
        return;

    /*   ,   */
    if (!CPLD_PowerGood())
        {
        f_RunState = DAC_ST_STOPPED;
        DIAG_CLEAR_FLAG(DIAG_DEVSTATE_DAC_STARTED);
#if DEBUG_LED_DAC_RUNNING
        DEBUG_LED_OFF();
#endif
        return;
        }

    /*      */
    /*     ,     */
    /* :        ,
        ..   USB    DAC_BLK_SIZE_B , 
             */
    if ((f_NeedBlocks == 0) || (f_BufState[f_NeedBlocks - 1] != BS_FULL))
        return;

    if (!SPI_OpenForDAC())
        {
        f_RunState = DAC_ST_STOPPED;
        DIAG_CLEAR_FLAG(DIAG_DEVSTATE_DAC_STARTED);
#if DEBUG_LED_DAC_RUNNING
        DEBUG_LED_OFF();
#endif
        return;
        }

    f_SendIdx = 0;
    f_StopTailLen = 0;
    f_StopRepeatPtr = f_StopBuf;
    f_LoopStart = g_Buf.DAC.asFrame;

    /*      DMA */
    if (f_LoopSamples)
        { /* ========   ======== */
        unsigned int chunk1_size, chunk2_size;
        volatile unsigned int *chunk1_ptr, *chunk2_ptr;
        unsigned int max_offset;

        /*        */

        /*     DAC_MIN_DMA_SAMPLES,     -
             */
        if (f_LoopSamples < DAC_MIN_DMA_SAMPLES)
            {
            unsigned int add_count = f_LoopSamples;
            register volatile unsigned int* p;
            while (f_LoopSamples + add_count < DAC_MIN_DMA_SAMPLES)
                {
                add_count += f_LoopSamples;
                }
            /*       */
            for (p = f_LoopStart; p < f_LoopStart + add_count; p++)
                {
                p[f_LoopSamples] = p[0];
                }
            f_LoopSamples += add_count;
            }

        /*       ,   , 
               ( DAC_BUFFER_SIZE_DW_TOTAL  DAC_BUFFER_SIZE_DW) */
        max_offset = f_LoopSamples - DAC_MIN_DMA_SAMPLES;
        if (f_StartOffset > max_offset)
            {
            unsigned int k;
            for (k = 0; k < DAC_MIN_DMA_SAMPLES; k++)
                {
                f_LoopStart[f_LoopSamples + k] = f_LoopStart[k];
                }
            f_LoopStart += (f_StartOffset - max_offset);
            f_StartOffset = max_offset;
            }

        chunk1_ptr = &f_LoopStart[f_StartOffset];
        f_LastCountCur = chunk1_size = f_LoopSamples - f_StartOffset;
        chunk2_ptr = f_LoopStart;
        f_LastCountNext = chunk2_size = f_LoopSamples;
        if (f_UseAutoStop)
            { /*           */
            unsigned int k, tail_len = 0, stop_sample;
            volatile unsigned int *tail_ptr;
            if (f_SamplesLeft <= chunk1_size)
                { /*      .  . */
                f_LastCountCur = chunk1_size = f_SamplesLeft;
                f_LastCountNext = 0;
                f_SamplesLeft = 0;
                tail_len = chunk1_size;
                tail_ptr = chunk1_ptr;
                if (chunk1_size < DAC_MIN_DMA_SAMPLES)
                    {
                    chunk1_ptr = f_StopBuf;
                    chunk1_size = DAC_MIN_DMA_SAMPLES;
                    }
                }
            else
                { /*    */
                f_SamplesLeft -= chunk1_size;
                if (f_SamplesLeft < chunk2_size)
                    f_LastCountNext = chunk2_size = f_SamplesLeft;
                tail_len = f_SamplesLeft % f_LoopSamples;
                tail_ptr = chunk2_ptr;
                }
            /*  f_SamplesLeft   ,     */

            /*      :
                 ( f_SamplesLeft % f_LoopSamples)    
                 DMA.   ,     
                .
             */
            stop_sample = (f_UseStopConst)
                ? f_StopConst
                : tail_ptr[((tail_len) ? tail_len : f_LoopSamples) - 1];
            if ((tail_len > 0) && (tail_len < DAC_MIN_DMA_SAMPLES))
                {
                /*  tail_len       */
                register volatile unsigned int *src = tail_ptr;
                /*        */
                if (f_SamplesLeft > 0)
                    f_SamplesLeft -= tail_len;
                f_StopTailLen = tail_len;
                while (tail_len--)
                    {
                    *f_StopRepeatPtr++ = *src++;
                    }
                /*    while  f_StopRepeatPtr = f_StopBuf + f_StopTailLen
                         - (c ). */
                }

            /*  -       */
            for (k = 0; k < DAC_MIN_DMA_SAMPLES; k++)
                {
                f_StopRepeatPtr[k] = stop_sample;
                }

            if (f_SamplesLeft == 0)
                { /*    --   */
                chunk2_ptr = (chunk1_ptr != f_StopBuf) ? f_StopBuf : f_StopRepeatPtr;
                chunk2_size = DAC_MIN_DMA_SAMPLES;
                }
            } /* ... f_UseAutoStop */
        AT91C_BASE_SPI->SPI_TPR = (unsigned int)chunk1_ptr;
        AT91C_BASE_SPI->SPI_TCR = chunk1_size * 2;
        AT91C_BASE_SPI->SPI_TNPR = (unsigned int)chunk2_ptr;
        AT91C_BASE_SPI->SPI_TNCR = chunk2_size * 2;
        } /* ... f_LoopSamples */
    else
        { /* ========   ======== */
        /*      2  (    ) */
        AT91C_BASE_SPI->SPI_TPR = (unsigned int)g_Buf.DAC.asRing[0];
        AT91C_BASE_SPI->SPI_TCR = DAC_BLK_SIZE_W;
        AT91C_BASE_SPI->SPI_TNPR = (unsigned int)g_Buf.DAC.asRing[1];
        AT91C_BASE_SPI->SPI_TNCR = DAC_BLK_SIZE_W;
        f_LastCountCur = f_LastCountNext = DAC_BLK_SIZE_DW;
        if (f_UseAutoStop)
            {
            unsigned int tail_len = f_SamplesLeft % DAC_BLK_SIZE_DW;
            /*          DMA,
                 ""         */
            if ((tail_len > 0) && (tail_len < DAC_MIN_DMA_SAMPLES))
                {
                /*        */
                f_SamplesLeft -= tail_len;
                f_StopTailLen = tail_len;
                /* !   "",      
                      (   ).     
                     (  f_StopConst ,  f_UseStopConst == 0,
                      .
                        . */
                }
            /*    (  DMA) */
            f_SamplesLeft -= DAC_BLK_SIZE_DW;
            } /* ... f_UseAutoStop */
        } /* ... !f_LoopSamples */

    /*   .
         (    f_SamplesLeft)  :

         ,     f_StopBuf (     ""
       ),  -       f_StopRepeatPtr,
           f_SamplesLeft == 0   2 .
       (  -     f_StopBuf,   -  
        f_StopRepeatPtr).

         ,        ,  
       f_StopBuf = f_StopRepeatPtr   .
         -     f_StopBuf ( ),  - 
            .

       ,  f_AutoStopDownCount      2.
       ( 1 + (f_StopTailLen != 0) ).
    */
    f_AutoStopDownCount = 2;

    /*   " " */
    f_RunState = DAC_ST_STREAMING;

    f_LastDMAPtr = (unsigned int*)AT91C_BASE_SPI->SPI_TNPR;
    SPI_StartDMATransmit(1);

    ok = CPLD_CFG_SetConfigBits(
         CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE,
         CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE);
    /*    */
    if (ok)
        ok = CPLD_DAC_Async(CPLD_DACCTL_ASYNC_PRELOAD);
    if (ok)
        ok = CPLD_DAC_Start(f_saved_SyncWithADC, f_saved_RateDiv);
    if (!ok)
        { /*  :    SPI ,      SPI */
        CPLD_CFG_SetConfigBits((CPLD_CONFIG_SPI_MASTER_ENABLE | CPLD_CONFIG_DAC_ENABLE), 0);
        SPI_Close();
        f_RunState = DAC_ST_STOPPED;
        DIAG_CLEAR_FLAG(DIAG_DEVSTATE_DAC_STARTED);
#if DEBUG_LED_DAC_RUNNING
        DEBUG_LED_OFF();
#endif
        return;
        }
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
