/*================================================================================================*
 *  EEPROM
 *================================================================================================*/

#include "eeprom.h"
#include "spi.h"
#include "timers.h"
#include "wdt.h"

/*================================================================================================*/
#define EECMD_WRSR          0x01    /* write status reg */
#define EECMD_WRITE         0x02    /* write eeprom */
#define EECMD_READ          0x03    /* read eeprom */
#define EECMD_WRDI          0x04    /* write disable */
#define EECMD_RDSR          0x05    /* read status reg */
#define EECMD_WREN          0x06    /* write enable */
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
unsigned char EEPROM_GetChipState   /*    EEPROM */
    (void)                          /*     0xFF   */
    {
    if (!SPI_IsOpenAs(SPI_MODE_MASTER))
        return 0xFF;

    /*    EEPROM */
    SPI_Xfer(EECMD_RDSR, SPI_CONT);
    return SPI_Xfer(0, SPI_LAST);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int EEPROM_Protect                  /*      EEPROM */
    (
    unsigned int NewState,          /* ()  -   (EEPROM_WP_...) */
    unsigned int ChipTimeout        /* ()  -    ,  (0 = ) */
    )                               /*  1  , 0   */
    {
    unsigned char status;

    if (!SPI_IsOpenAs(SPI_MODE_MASTER))
        return 0;

    /*    EEPROM */
    SPI_Xfer(EECMD_RDSR, SPI_CONT);
    status = SPI_Xfer(0, SPI_LAST) & 0xFF;
    if (status & EEPROM_ST_BUSY) /* bit0 = nRDY, EEPROM   */
        return 0;

    /*    */
    SPI_Xfer(EECMD_WREN, SPI_LAST);
    SPI_Xfer(EECMD_WRSR, SPI_CONT);
    SPI_Xfer(NewState & EEPROM_WP_ALL, SPI_LAST);
    if (ChipTimeout)
        {
        TIMER_StartTimeout(ChipTimeout);
        do
            {
            WDT_Restart();
            SPI_Xfer(EECMD_RDSR, SPI_CONT);
            status = SPI_Xfer(0, SPI_LAST) & 0xFF;
            }
        while (!TIMER_TimeoutExpired() && (status & EEPROM_ST_BUSY));
        if (status & EEPROM_ST_BUSY)
            return 0;
        }
    else
        {
        WDT_Restart();
        do
            {
            SPI_Xfer(EECMD_RDSR, SPI_CONT);
            status = SPI_Xfer(0, SPI_LAST) & 0xFF;
            }
        while (status & EEPROM_ST_BUSY);
        }

    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int EEPROM_Read                     /*    EEPROM */
    (
    unsigned int StartAddr,         /* ()  -    EEPROM */
    unsigned int NumBytes,          /* ()  -    */
    unsigned char *pBuffer          /* () - - */
    )                               /*  1  , 0   */
    {
    unsigned char status;

    if (!SPI_IsOpenAs(SPI_MODE_MASTER))
        return 0;
    if ((StartAddr >= EEPROM_SIZE) || (NumBytes == 0) || (NumBytes > EEPROM_SIZE))
        return 0;
    if (!pBuffer)
        return 0;

    /*    EEPROM */
    SPI_Xfer(EECMD_RDSR, SPI_CONT);
    status = SPI_Xfer(0, SPI_LAST) & 0xFF;
    if (status & EEPROM_ST_BUSY)
        return 0;

    SPI_Xfer(EECMD_READ, SPI_CONT);
#if EEPROM_SIZE >= 0x100
    SPI_Xfer((StartAddr >> 8) & 0xFF, SPI_CONT);
#endif
    SPI_Xfer(StartAddr & 0xFF, SPI_CONT);
    while (NumBytes--)
        {
        *pBuffer++ = SPI_Xfer(0, (NumBytes ? SPI_CONT : SPI_LAST)) & 0xFF;
        }

    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int EEPROM_Write                    /*    EEPROM */
    (
    unsigned int StartAddr,         /* ()  -    EEPROM */
    unsigned int NumBytes,          /* ()  -    */
    const unsigned char *pBuffer,   /* ()  - - */
    unsigned int ChipTimeout        /* ()  -    ,  (0 = ) */
    )                               /*  1  , 0   */
    {
    unsigned char status;
    unsigned int BlockSize;

    if (NumBytes == 0)
        return 1;
    if (!SPI_IsOpenAs(SPI_MODE_MASTER))
        return 0;
    if ((StartAddr >= EEPROM_SIZE)
        || (NumBytes > EEPROM_SIZE)
        || (StartAddr + NumBytes > EEPROM_SIZE))
        return 0;
    if (!pBuffer)
        return 0;

    /*    EEPROM */
    SPI_Xfer(EECMD_RDSR, SPI_CONT);
    status = SPI_Xfer(0, SPI_LAST) & 0xFF;
    if (status & EEPROM_ST_BUSY) /* bit0 = nRDY, EEPROM   */
        return 0;

    /*    */
    while (NumBytes > 0)
        {
        /*       :    
             , ..   EEPROM_PAGE_SIZE */
        BlockSize = EEPROM_PAGE_SIZE - (StartAddr % EEPROM_PAGE_SIZE);
        if (BlockSize > NumBytes)
            BlockSize = NumBytes;
        /*  :   WREN  WRITE */
        SPI_Xfer(EECMD_WREN, SPI_LAST);
        SPI_Xfer(EECMD_WRITE, SPI_CONT);
#if EEPROM_SIZE >= 0x100
        SPI_Xfer((StartAddr >> 8) & 0xFF, SPI_CONT);
#endif
        SPI_Xfer(StartAddr & 0xFF, SPI_CONT);
        StartAddr += BlockSize;
        NumBytes -= BlockSize;
        /*   */
        while (BlockSize--)
            {
            SPI_Xfer(*pBuffer++, (BlockSize ? SPI_CONT : SPI_LAST));
            }
        /*    */
        if (ChipTimeout)
            {
            TIMER_StartTimeout(ChipTimeout);
            do
                {
                WDT_Restart();
                SPI_Xfer(EECMD_RDSR, SPI_CONT);
                status = SPI_Xfer(0, SPI_LAST) & 0xFF;
                }
            while (!TIMER_TimeoutExpired() && (status & EEPROM_ST_BUSY));
            if (status & EEPROM_ST_BUSY)
                return 0;
            }
        else
            {
            WDT_Restart();
            do
                {
                SPI_Xfer(EECMD_RDSR, SPI_CONT);
                status = SPI_Xfer(0, SPI_LAST) & 0xFF;
                }
            while (status & EEPROM_ST_BUSY);
            }
        } /* while (NumBytes > 0) */

    return 1;
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
