#include "main.h"
#include "cpld.h"
#include "adc.h"
#include "dac.h"
#include "spi.h"
#include "pio.h"
#include "debug.h"
#include "USB_USBMain.h"
#include "USBBackground.h"
#include "nvram.h"
#include "timers.h"
#include "wdt.h"
#include "bootrom.h"
#include "bootflash.h"

/*================================================================================================*/
/*  ,       USB */
unsigned int g_MainFlag = 0;

const char ModuleName[7] = MODULE_NAME;

t_DiagVars g_Diag;

typedef struct
    {
    unsigned char Signature[4];     /*  "VER." */
    unsigned char VerMajor[2];      /*     */
    unsigned char VerMinor[2];      /*     */
    unsigned char Date[8];          /*    "YYYYMMDD" */
    }
    t_BinInfoBlock;
/*  ,    __DATE__ "Dec  1 2000"  "20001201" */
#define BUILD_MONTH ( \
    ((__DATE__)[2] == 'n') ? /* Jan or Jun */       \
        (((__DATE__)[1] == 'a') ? "01" : "06") :    \
    ((__DATE__)[2] == 'b') ? "02" :                 \
    ((__DATE__)[2] == 'r') ? /* Mar or Apr */       \
        (((__DATE__)[1] == 'a') ? "03" : "04") :    \
    ((__DATE__)[2] == 'y') ? "05" :                 \
    ((__DATE__)[2] == 'l') ? "07" :                 \
    ((__DATE__)[2] == 'g') ? "08" :                 \
    ((__DATE__)[2] == 'p') ? "09" :                 \
    ((__DATE__)[2] == 't') ? "10" :                 \
    ((__DATE__)[2] == 'v') ? "11" :                 \
    ((__DATE__)[2] == 'c') ? "12" :                 \
    "")

const __root t_VersionInfo VersionInfo =
    {
    VER_MINOR, VER_MAJOR, __DATE__"\0\0\0"
    };

#pragma location = "TIMESTAMP_SECTION"
const __root t_BinInfoBlock BinInfoBlock =
    {
    "VER.",
    { ((VER_MAJOR / 10) % 10) + '0', (VER_MAJOR % 10) + '0' },
    { ((VER_MINOR / 10) % 10) + '0', (VER_MINOR % 10) + '0' },
    { /* Date */
        (__DATE__)[7], (__DATE__)[8], (__DATE__)[9], (__DATE__)[10], /* Year */
        BUILD_MONTH[0], BUILD_MONTH[1],                              /* Month */
        ((__DATE__)[4] == ' ') ? '0' : (__DATE__)[4], (__DATE__)[5]  /* Day */
    }
    };
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
int SetLowPowerMode                 /*     */
    (
    int LowPower                    /* ()  -   (1)   (0)  */
    )                               /*  1  , 0   */
    {
    if (LowPower)
        { /* Enter low power mode */
        /*    ,    */
        ADC_Stop();
        DAC_Stop(1);
        /*   */
        CPLD_PowerOff();
        }
    else
        { /* Leave low power mode */
        /*   ,     (   ) */
        if (CPLD_PowerGood())
            return 1;
        /*   */
        if (!CPLD_PowerOn(CPLD_POWER_ON_TIMEOUT))
            return 0;
        }
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void main()
    {
    int have_rq;

    /*    bootrom   usb device request,    */
    have_rq = DeviceConfigured() && BOOTROM_GetRequest(&CurrentUSBRequest);
    BOOTROM_Clear(); /*     */

    BOOTFLASH_UpdateBootrom();

    TIMER_Init();
    ADC_Init();
    SPI_Init();
    DAC_Init();

    /*  USART   */
    CPLD_InitUSART();

    /*  USB */
    if (!have_rq)
        { /* no bootrom, USB not configured */
        ConnectUSB();
        while (!USBConfigured)
            {
            USBBG_Worker(); /*   watchdog */
            }
        }
    else
        { /* after bootrom, USB already configured */
        ConnectUSBFromBoot();
        USBConfigured = true;
        g_TIMER_BlinkFlag = 1;
        }

    /*  ,        */
    SetLowPowerMode(0);

    /*         */
    CPLD_Initialize(
        (0 & CPLD_CONFIG_TTLOUT_ENABLE) |
        (0 & CPLD_CONFIG_DAC_ENABLE) |
        (0 & CPLD_CONFIG_SPI_MASTER_ENABLE)
        );

    NVRAM_Init();

    if (have_rq)
        {
        if (CurrentUSBRequest.bRequest < VDRQ_COUNT)
            {
            __disable_interrupt();
            VendorDeviceRequest[CurrentUSBRequest.bRequest]();
            __enable_interrupt();
            }
        }

    for(;;)
        {
        USBBG_Worker(); /*   watchdog */
        ADC_Worker();
        DAC_Worker();

        if (g_MainFlag & MAIN_JUMP_TO_BOOTROM)
            {
            g_MainFlag &= ~MAIN_JUMP_TO_BOOTROM;
            if (!BOOTROM_Present())
                {
                USBBG_RequestError();
                }
            else
                {
                // SetLowPowerMode(1); /*      */
                __disable_interrupt();
                BOOTROM_SetCommand(BOOTCMD_ENTER_BOOTROM);
                BOOTROM_SetRequest(&CurrentUSBRequest);
                AT91C_BASE_MC->MC_RCR = AT91C_MC_RCB; /*  REMAP */
                AT91C_BASE_RSTC->RSTC_RCR = 0xA5000000 | AT91C_RSTC_PROCRST; /* Reset core only */
                /* AT91C_RSTC_PERRST  ,  USB     */
                for (;;);
                }
            }

        if ((g_MainFlag & MAIN_RESET_DEVICE) && !USBBG_TaskRunning())
            {
            g_MainFlag &= ~MAIN_RESET_DEVICE;
            AT91C_BASE_RSTC->RSTC_RCR = 0xA5000000 |
                AT91C_RSTC_PROCRST | AT91C_RSTC_PERRST | AT91C_RSTC_EXTRST;
            for (;;);
            }

#if MAIN_SLEEP_ON_USB_SUSPEND
        if (g_MainFlag & MAIN_ENTER_SLEEP_MODE)
            {
            g_MainFlag &= ~MAIN_ENTER_SLEEP_MODE;
            SetLowPowerMode(1);
            }

        if (g_MainFlag & MAIN_EXIT_SLEEP_MODE)
            {
            g_MainFlag &= ~MAIN_EXIT_SLEEP_MODE;
            SetLowPowerMode(0);
            }
#endif
        }
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
