/*================================================================================================*
 *    ( )
 *================================================================================================*/

#include "nvram.h"
#include "eeprom.h"
#include "spi.h"
#include "chipio.h"
#include "clk.h"
#include "wdt.h"
#include <string.h>
#include <intrinsics.h>

/*================================================================================================*/
#define DEVDESC_CACHE       1       /*     RAM */
#define USERAREA_CACHE      1       /*      RAM */
#define USERAREA_IN_FLASH   (!EEPROM_8KBIT) /* 1 =    Flash; 0 =  EEPROM */

#define DEVDESC_EE_ADDR     0       /*      EEPROM */
#define EEPROM_TIMEOUT      100     /*     EEPROM,  */

#if DEVDESC_EE_ADDR + NVRAM_DEVDESC_SIZE > EEPROM_SIZE
#error "Device descriptor does not fit in EEPROM"
#endif

#if USERAREA_IN_FLASH
/* -- USERAREA in FLASH -- */
#define FLASH_PAGE_SIZE     256     /*   FLASH   */
#define USERAREA_NUM_PAGES  ((NVRAM_USERAREA_SIZE + FLASH_PAGE_SIZE - 1) / FLASH_PAGE_SIZE)
#if !USERAREA_CACHE
#error "USERAREA_IN_FLASH requires USERAREA_CACHE to be enabled"
#endif
#if (NVRAM_USERAREA_SIZE % FLASH_PAGE_SIZE)
#error "User area must occupy a whole number of flash pages"
#endif
#else
/* -- USERAREA in EEPROM -- */
#define USERAREA_EE_ADDR  EEPROM_SIZE - NVRAM_USERAREA_SIZE /*   userarea  EEPROM */
#if USERAREA_EE_ADDR < 0
#error "User area does not fit in EEPROM"
#endif
#endif
/*================================================================================================*/

/*================================================================================================*/
static int f_WriteEnable = 0;
static int f_InitOk = 0;
#if DEVDESC_CACHE
static unsigned char f_DescriptorCache[NVRAM_DEVDESC_SIZE];
#endif
#if USERAREA_CACHE
static unsigned char f_UserAreaCache[NVRAM_USERAREA_SIZE];
#endif

#if USERAREA_IN_FLASH
static int f_UserFlashUnsaved[USERAREA_NUM_PAGES];
#pragma location = "USER_FLASH_SECTION"
__no_init unsigned char UserFlash[NVRAM_USERAREA_SIZE];
#endif
/*================================================================================================*/

/*================================================================================================*/
#if USERAREA_IN_FLASH
/*------------------------------------------------------------------------------------------------*/
#pragma diag_suppress=Ta021,Ta023
static __ramfunc void f_ProgramFlash
    (unsigned int LinearAddress)
    {
    __disable_interrupt();
    WDT_Restart();
    while (!(AT91C_BASE_MC->MC_FSR & AT91C_MC_FRDY));
    AT91C_BASE_MC->MC_FCR = AT91C_MC_CORRECT_KEY | AT91C_MC_FCMD_START_PROG |
        (AT91C_MC_PAGEN & LinearAddress);
    while (!(AT91C_BASE_MC->MC_FSR & AT91C_MC_FRDY));
    __enable_interrupt();
    }
#pragma diag_default=Ta021,Ta023
/*------------------------------------------------------------------------------------------------*/
#endif

/*------------------------------------------------------------------------------------------------*/
void NVRAM_Init                     /*  */
    (void)
    {
    f_InitOk = 0;
    /*  -    */
#if DEVDESC_CACHE || USERAREA_CACHE
    if (!SPI_OpenForEEPROM())
        return;
    #if DEVDESC_CACHE
    if (!EEPROM_Read(DEVDESC_EE_ADDR, NVRAM_DEVDESC_SIZE, f_DescriptorCache))
        {
        SPI_Close();
        return;
        }
    #endif
    #if USERAREA_CACHE
        #if USERAREA_IN_FLASH
        memcpy(f_UserAreaCache, UserFlash, NVRAM_USERAREA_SIZE);
        #else
        if (!EEPROM_Read(USERAREA_EE_ADDR, NVRAM_USERAREA_SIZE, f_UserAreaCache))
            {
            SPI_Close();
            return;
            }
        #endif
    #endif
    SPI_Close();
#endif
    /*   flash,      */
#if USERAREA_IN_FLASH
    memset(f_UserFlashUnsaved, 0, sizeof(f_UserFlashUnsaved));
#endif
    f_InitOk = 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void NVRAM_FlushBuffers             /*   ,  ,  NVRAM */
    (void)
    {
#if USERAREA_IN_FLASH
    /*   flash,      */
    unsigned int *pSrc, *pDest;
    int k;

    if (!f_InitOk) return;
    pSrc = (unsigned int*)f_UserAreaCache;
    pDest = (unsigned int*)UserFlash;
    for (k = 0; k < USERAREA_NUM_PAGES; k++)
        {
        if (f_UserFlashUnsaved[k])
            {
            int i;
            /*      flash *32- * */
            for (i = 0; i < FLASH_PAGE_SIZE / 4; i++)
                {
                pDest[i] = pSrc[i];
                }
            /*  flash (    RAM) */
            f_ProgramFlash((unsigned int)pDest);
            /*    */
            f_UserFlashUnsaved[k] = 0;
            }
        pSrc += FLASH_PAGE_SIZE / 4;
        pDest += FLASH_PAGE_SIZE / 4;
        }
#endif
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int NVRAM_ReadDeviceDescriptor      /*    */
    (
    unsigned int ByteOffset,        /* ()  -     */
    unsigned char *pBuffer,         /* () -     */
    unsigned int NumBytes           /* ()  -    */
    )                               /*  1  , 0   */
    {
    int ok;

    if (!f_InitOk) return 0;
    if (!pBuffer) return 0;
    if (NumBytes == 0) return 1;
    if ((ByteOffset >= NVRAM_DEVDESC_SIZE)
        || (NumBytes > NVRAM_DEVDESC_SIZE)
        || (ByteOffset + NumBytes > NVRAM_DEVDESC_SIZE))
        return 0;
#if DEVDESC_CACHE
    /*   ,      */
    memcpy(pBuffer, (char*)f_DescriptorCache + ByteOffset, NumBytes);
    ok = 1;
#else
    /*    -    EEPROM */
    ok = SPI_OpenForEEPROM();
    if (ok)
        {
        ok = EEPROM_Read(DEVDESC_EE_ADDR + ByteOffset, NumBytes, pBuffer);
        SPI_Close();
        }
#endif
    return ok;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int NVRAM_WriteDeviceDescriptor     /*    */
    (
    unsigned int ByteOffset,        /* ()  -     */
    const unsigned char *pBuffer,   /* ()  -   */
    unsigned int NumBytes           /* ()  -    */
    )                               /*  1  , 0   */
    {
    int ok;

    if (!f_InitOk) return 0;
    if (!pBuffer) return 0;
    if (!f_WriteEnable) return 0;
    if (NumBytes == 0) return 1;
    if ((ByteOffset >= NVRAM_DEVDESC_SIZE)
        || (NumBytes > NVRAM_DEVDESC_SIZE)
        || (ByteOffset + NumBytes > NVRAM_DEVDESC_SIZE))
        return 0;

    /*  EEPROM */
    if (!SPI_OpenForEEPROM())
        return 0;
    /*     */
    ok = EEPROM_Protect(EEPROM_WP_NONE, EEPROM_TIMEOUT);
    /*   */
    if (ok)
        ok = EEPROM_Write(DEVDESC_EE_ADDR + ByteOffset, NumBytes, pBuffer, EEPROM_TIMEOUT);
    /*     */
    if (ok)
        ok = EEPROM_Protect(EEPROM_WP_ALL, EEPROM_TIMEOUT);
    /*  EEPROM */
    SPI_Close();
#if DEVDESC_CACHE
    if (ok)
        memcpy((char*)f_DescriptorCache + ByteOffset, pBuffer, NumBytes);
#endif
    return ok;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int NVRAM_ReadUserArea              /*    NVRAM */
    (
    unsigned int ByteOffset,        /* ()  -     */
    unsigned char *pBuffer,         /* () -     */
    unsigned int NumBytes           /* ()  -    */
    )                               /*  1  , 0   */
    {
    int ok;

    if (!f_InitOk) return 0;
    if (!pBuffer) return 0;
    if (NumBytes == 0) return 1;
    if ((ByteOffset >= NVRAM_USERAREA_SIZE)
        || (NumBytes > NVRAM_USERAREA_SIZE)
        || (ByteOffset + NumBytes > NVRAM_USERAREA_SIZE))
        return 0;
#if USERAREA_CACHE
    /*   ,      */
    memcpy(pBuffer, (char*)f_UserAreaCache + ByteOffset, NumBytes);
    ok = 1;
#elif USERAREA_IN_FLASH
    /*    -    FLASH */
    memcpy(pBuffer, (char*)UserFlash + ByteOffset, NumBytes);
    ok = 1;
#else /* USERAREA in EEPROM */
    /*    -    EEPROM */
    ok = SPI_OpenForEEPROM();
    if (ok)
        {
        ok = EEPROM_Read(USERAREA_EE_ADDR + ByteOffset, NumBytes, pBuffer);
        SPI_Close();
        }
#endif
    return ok;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int NVRAM_WriteUserArea             /*     NVRAM */
    (
    unsigned int ByteOffset,        /* ()  -     */
    const unsigned char *pBuffer,   /* ()  -   */
    unsigned int NumBytes           /* ()  -    */
    )                               /*  1  , 0   */
    {
    int ok;

    if (!f_InitOk) return 0;
    if (!pBuffer) return 0;
    if (!f_WriteEnable) return 0;
    if (NumBytes == 0) return 1;
    if ((ByteOffset >= NVRAM_USERAREA_SIZE)
        || (NumBytes > NVRAM_USERAREA_SIZE)
        || (ByteOffset + NumBytes > NVRAM_USERAREA_SIZE))
        return 0;
#if USERAREA_IN_FLASH
    /*     (   NVRAM_FlushBuffers()) */
    memcpy((char*)f_UserAreaCache + ByteOffset, pBuffer, NumBytes);
    /*    flash   */
    for (ok = ByteOffset / FLASH_PAGE_SIZE;
         ok <= (ByteOffset + NumBytes - 1) / FLASH_PAGE_SIZE;
         ok++)
        {
        f_UserFlashUnsaved[ok] = 1;
        }
    ok = 1;
#else /* USERAREA in EEPROM */
    /*  EEPROM */
    if (!SPI_OpenForEEPROM())
        return 0;
    /*     */
    ok = EEPROM_Protect(EEPROM_WP_NONE, EEPROM_TIMEOUT);
    /*   */
    if (ok)
        ok = EEPROM_Write(USERAREA_EE_ADDR + ByteOffset, NumBytes, pBuffer, EEPROM_TIMEOUT);
    /*     */
    if (ok)
        ok = EEPROM_Protect(EEPROM_WP_ALL, EEPROM_TIMEOUT);
    /*  EEPROM */
    SPI_Close();
#if USERAREA_CACHE
    if (ok)
        memcpy((char*)f_UserAreaCache + ByteOffset, pBuffer, NumBytes);
#endif
#endif
    return ok;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void NVRAM_WriteEnable              /*    */
    (
    int Enable                      /* ()  - 1 =   , 0 =   */
    )
    {
    f_WriteEnable = Enable;
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
