/*================================================================================================*
 *      USB (control channel)
 *================================================================================================*/

#include "USBBackground.h"
#include "USBHardware.h"
#include "wdt.h"

#define BYPASS  0

/*================================================================================================*/
/*   */
enum
    {
    USBBG_ST_IDLE,          /*  */
    USBBG_ST_READ_USB,      /*    */
    USBBG_ST_WRITE_PREPARE, /*  ( callback)    */
    USBBG_ST_WRITE_USB,     /*   */
    USBBG_ST_WRITE_WAIT,    /*      */
    USBBG_ST_FIN_OK,        /*      (OK) */
    USBBG_ST_FIN_ERROR      /*    */
    };
/*================================================================================================*/

/*================================================================================================*/
static struct                       /*       */
    {
    unsigned int State;             /*   (USBBG_ST_...) */
    unsigned char *pBuf;            /*    / */
    unsigned int UserParam;         /*   -  */
    unsigned int RequestSize;       /* ,    */
    unsigned int DataOffset;        /*  /     */
    t_USBBG_Callback Callback;      /*   callback   */
    }
    f_Ctl = { USBBG_ST_IDLE, NULL, 0, 0, 0, NULL };
static unsigned char f_DefaultBuffer[USBBG_DEFBUF_SIZE];
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
int USBBG_Begin                     /*          */
    (
    int Direction,                  /* ()  - USBBG_READ  USBBG_WRITE */
    unsigned int UserParam,         /* ()  -   -  */
    unsigned int USBRequestSize,    /* ()  -   */
    t_USBBG_Callback Callback       /* ()  -   callback   */
    )                               /*  1    0   */
    {
    return USBBG_BeginWithBuffer(Direction, UserParam, USBRequestSize, Callback,
                                 f_DefaultBuffer, sizeof(f_DefaultBuffer));
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int USBBG_BeginWithBuffer           /*         */
    (
    int Direction,                  /* ()  - USBBG_READ  USBBG_WRITE */
    unsigned int UserParam,         /* ()  -   -  */
    unsigned int USBRequestSize,    /* ()  -   */
    t_USBBG_Callback Callback,      /* ()  -   callback   */
    void *pBuffer,                  /* ()  -    */
    unsigned int BufferSize         /* ()  -   (    USBRequestSize) */
    )                               /*  1    0   */
    {
    if (f_Ctl.State != USBBG_ST_IDLE)
        return 0;
    if (!pBuffer)
        return 0;
    if (USBRequestSize > BufferSize)
        { /* ,     */
        if (Direction == USBBG_WRITE) return 0; /*   --  */
        USBRequestSize = BufferSize; /*   ,     */
        }
    f_Ctl.pBuf = pBuffer;
    f_Ctl.UserParam = UserParam;
    f_Ctl.RequestSize = USBRequestSize;
    f_Ctl.DataOffset = 0;
    f_Ctl.Callback = Callback;
    f_Ctl.State = (Direction == USBBG_WRITE) ? USBBG_ST_READ_USB : USBBG_ST_WRITE_PREPARE;
    Disable_USB_Interrupts(1 << CONTROL_ENDPOINT);
#if BYPASS
    if (!Callback)
        while (USBBG_TaskRunning()) USBBG_Worker();
#endif
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int USBBG_RequestError              /*    ( stall) */
    (void)                          /*  1    0   */
    {
    if (f_Ctl.State != USBBG_ST_IDLE)
        return 0;
    /*  stall */
    Disable_USB_Interrupts(1 << CONTROL_ENDPOINT);
    Udp_ep_set_flag(CONTROL_ENDPOINT, AT91C_UDP_FORCESTALL);
    f_Ctl.State = USBBG_ST_FIN_ERROR;
#if BYPASS
    while (USBBG_TaskRunning()) USBBG_Worker();
#endif
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int USBBG_RequestOK                 /*    (  ) */
    (void)                          /*  1    0   */
    {
    if (f_Ctl.State != USBBG_ST_IDLE)
        return 0;
    Disable_USB_Interrupts(1 << CONTROL_ENDPOINT);
    /*     */
    Udp_ep_set_flag(CONTROL_ENDPOINT, AT91C_UDP_TXPKTRDY);
    f_Ctl.State = USBBG_ST_FIN_OK;
#if BYPASS
    while (USBBG_TaskRunning()) USBBG_Worker();
#endif
    return 1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int USBBG_TaskRunning               /* ,       */
    (void)
    {
    return (f_Ctl.State != USBBG_ST_IDLE);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void USBBG_Worker                   /*      main() */
    (void)
    {
    int pkt_size, bytes_left, again;
    static int last_pkt_size;

    WDT_Restart(); /*  watchdog */

    do switch (again = 0, f_Ctl.State)
        {
        case USBBG_ST_READ_USB:
            /* ----     callback   ---- */
            bytes_left = f_Ctl.RequestSize - f_Ctl.DataOffset;
            if (bytes_left > 0)
                {
                /*    */
                if (!(AT91C_BASE_UDP->UDP_CSR[CONTROL_ENDPOINT] & AT91C_UDP_RX_DATA_BK0))
                    break;
                /*    */
                pkt_size = (AT91C_BASE_UDP->UDP_CSR[CONTROL_ENDPOINT] & AT91C_UDP_RXBYTECNT) >> 16;
                if (pkt_size > bytes_left)
                    pkt_size = bytes_left;
                USB_BEGIN_CRITICAL(1);
                reg2mem(&f_Ctl.pBuf[f_Ctl.DataOffset], &AT91C_BASE_UDP->UDP_FDR[CONTROL_ENDPOINT],
                        pkt_size);
                f_Ctl.DataOffset += pkt_size;
                Udp_ep_clr_flag(CONTROL_ENDPOINT, AT91C_UDP_RX_DATA_BK0);
                USB_END_CRITICAL(1);
                }
            else /*    -  callback    */
                {
                if (!f_Ctl.Callback ||
                    (f_Ctl.Callback(f_Ctl.UserParam, f_Ctl.pBuf, f_Ctl.RequestSize) >= 0))
                    { /*  -    */
                    Udp_ep_set_flag(CONTROL_ENDPOINT, AT91C_UDP_TXPKTRDY);
                    f_Ctl.State = USBBG_ST_FIN_OK;
                    }
                else
                    { /*  --   stall */
                    Udp_ep_set_flag(CONTROL_ENDPOINT, AT91C_UDP_FORCESTALL);
                    f_Ctl.State = USBBG_ST_FIN_ERROR;
                    }
                }
            break;
        case USBBG_ST_WRITE_PREPARE:
            /* ----  callback    ---- */
            bytes_left = (f_Ctl.Callback)
                ? f_Ctl.Callback(f_Ctl.UserParam, f_Ctl.pBuf, f_Ctl.RequestSize)
                : f_Ctl.RequestSize;
            if (bytes_left >= 0)
                {
                last_pkt_size = 0;
                /*    ,  ,    */
                if (bytes_left < f_Ctl.RequestSize)
                    f_Ctl.RequestSize = bytes_left;
                f_Ctl.State = USBBG_ST_WRITE_USB;
                again = 1;
                }
            else
                { /*  --   stall */
                Udp_ep_set_flag(CONTROL_ENDPOINT, AT91C_UDP_FORCESTALL);
                f_Ctl.State = USBBG_ST_FIN_ERROR;
                }
            break;
        case USBBG_ST_WRITE_USB:
            /* ----   ---- */
            bytes_left = f_Ctl.RequestSize - f_Ctl.DataOffset;
            if ((bytes_left > 0) || (last_pkt_size == CONTROL_ENDPOINT_SIZE))
                {
                pkt_size = (bytes_left < CONTROL_ENDPOINT_SIZE) ? bytes_left : CONTROL_ENDPOINT_SIZE;
                last_pkt_size = pkt_size;
                USB_BEGIN_CRITICAL(1);
                /*    FIFO */
                mem2reg(&AT91C_BASE_UDP->UDP_FDR[CONTROL_ENDPOINT], &f_Ctl.pBuf[f_Ctl.DataOffset],
                        pkt_size);
                f_Ctl.DataOffset += pkt_size;
                /*  ISOERROR  TXCOMP,    */
                Udp_ep_clr_flag(CONTROL_ENDPOINT, AT91C_UDP_ISOERROR);
                Udp_ep_clr_flag(CONTROL_ENDPOINT, AT91C_UDP_TXCOMP);
                /*   */
                Udp_ep_set_flag(CONTROL_ENDPOINT, AT91C_UDP_TXPKTRDY);
                USB_END_CRITICAL(1);
                f_Ctl.State = USBBG_ST_WRITE_WAIT;
                }
            else  /*   . */
                {
                Udp_ep_clr_flag(CONTROL_ENDPOINT, AT91C_UDP_TXCOMP);
                f_Ctl.State = USBBG_ST_IDLE;
                Clear_Usb_Interrupts(1 << CONTROL_ENDPOINT);
                Enable_USB_Interrupts(1 << CONTROL_ENDPOINT);
                }
            break;
        case USBBG_ST_WRITE_WAIT:
            /* ----   (   ) ---- */
            if (AT91C_BASE_UDP->UDP_CSR[CONTROL_ENDPOINT] & AT91C_UDP_TXCOMP)
                {
                //Udp_ep_clr_flag(CONTROL_ENDPOINT, AT91C_UDP_TXCOMP);
                f_Ctl.State = USBBG_ST_WRITE_USB;
                }
            else
            if (AT91C_BASE_UDP->UDP_CSR[CONTROL_ENDPOINT] & (AT91C_UDP_RX_DATA_BK0 | AT91C_UDP_RXSETUP))
                { /*     ,   */
                Udp_ep_clr_flag(CONTROL_ENDPOINT, AT91C_UDP_RX_DATA_BK0);
                f_Ctl.State = USBBG_ST_IDLE;
                Clear_Usb_Interrupts(1 << CONTROL_ENDPOINT);
                Enable_USB_Interrupts(1 << CONTROL_ENDPOINT);
                }
            break;
        case USBBG_ST_FIN_OK:
            /* ----     (OK) ---- */
            if (AT91C_BASE_UDP->UDP_CSR[CONTROL_ENDPOINT] & AT91C_UDP_TXCOMP)
                {
                Udp_ep_clr_flag(CONTROL_ENDPOINT, AT91C_UDP_TXCOMP);
                f_Ctl.State = USBBG_ST_IDLE;
                Clear_Usb_Interrupts(1 << CONTROL_ENDPOINT);
                Enable_USB_Interrupts(1 << CONTROL_ENDPOINT);
                }
            break;
        case USBBG_ST_FIN_ERROR:
            /* ----  stall ---- */
            if (AT91C_BASE_UDP->UDP_CSR[CONTROL_ENDPOINT] & AT91C_UDP_ISOERROR)
                {
                Udp_ep_clr_flag(CONTROL_ENDPOINT, AT91C_UDP_FORCESTALL | AT91C_UDP_ISOERROR);
                f_Ctl.State = USBBG_ST_IDLE;
                Clear_Usb_Interrupts(1 << CONTROL_ENDPOINT);
                Enable_USB_Interrupts(1 << CONTROL_ENDPOINT);
                }
            break;
        }
    while (again);
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
